unit dp_code;

// Dynamic Pascal, Copyright Lars Middendorf

interface

uses windows,classes,typinfo,inifiles,sysutils,contnrs,dp_scanner;

type

TBaseType=(btinteger,btenum,btboolean,btsingle,btstring,btrecord,btarray,btopenarray,btclass,btclassref,btmethod,btprocedure,btuntyped,btnil);

TTypeInfo=class;

TParamType=(ptnone,ptvalue,ptvar,ptconst);

TFieldVisibility=(fsprivate,fsprotected,fspublic,fspublished);

TMethodType=(mtprocedure,mtfunction,mtconstructor,mtdestructor,mtclassprocedure,mtclassfunction);

TSymbolFlag=(sfvirtual,sfnative,sfreference,sfimplicit,sftrueconst,sfmessage);
TSymbolFlags=set of TSymbolFlag;

TUnitInfo=class;

TRegister=(regnone,regeax,regedx,regecx,regebx,regedi,regesi);


TSymbolInfo=class
private
 fname:string;
 ftypeinfo:TTypeInfo;
 fparamtype:TParamType;
 foffset:Integer;
 fvmtindex:integer;
 fflags:TSymbolFlags;
 funitinfo:TUnitInfo;
 fvisibility:TFieldVisibility;
 fimplemented:boolean;
 fmethodptr:Pointer;
 fparent:TTypeInfo;
 fregister:TRegister;
 fconstdata:string;
 finterfacedecl:Boolean;
 fexecutemem:pointer;
 frefparam:TSymbolInfo;
 fmessageid:integer;
public
 property ConstData:string read fconstdata write fconstdata;
 property Flags:TSymbolFlags read fflags write fflags;
 property Implemented:Boolean read fimplemented write fimplemented;
 property InterfaceDecl:Boolean read finterfacedecl write finterfacedecl;
 property MessageID:Integer read fmessageid write fmessageid;
 property MethodPtr:Pointer read fmethodptr;
 property Name:String read fname;
 property Offset:Integer read foffset write foffset;
 property ParamType:TParamType read fparamtype write fparamtype;
 property Parent:TTypeInfo read fparent;
 property ParamRegister:TRegister read fregister;
 property RefParam:TSymbolInfo read frefparam write frefparam;
 property TypeInfo:TTypeInfo read ftypeinfo;
 property UnitInfo:TUnitInfo read funitinfo;
 property Visibility:TFieldVisibility read fvisibility;
 property VMTIndex:Integer read fvmtindex;
 constructor create(const n:string;t:TTypeInfo);
 destructor destroy;override;
end;

TTypeInfo=class
private
 fname:string;
 fshortname:shortstring;
 fbasetype:TBaseType;
 ffields:THashedStringList;
 fsize,finstsize:integer;
 fbasetypeptr:TTypeInfo;
 funitinfo:TUnitInfo;
 fmethodtype:TMethodType;
 fvmtsize:integer;
 fclassptr:TClass;
 fvmtptr:pointer;
 fdyntabptr:pointer;
 fparentvmt:pointer;
 finterfacedecl:Boolean;
 fclasstypeinfo:TTypeInfo;
 fusedparamreg:TRegister;
 fmin,fmax:integer;
 fresultptr,fmethod:boolean;
 frttdata:PTypeInfo;
 frttinfo:PPTypeInfo;
 unregister:boolean;
 function GetFieldCount:Integer;
 function GetField(i:integer):TSymbolInfo;
 procedure SetField(i:integer;v:TSymbolInfo);
 function MethodEquals(t2:TTypeInfo):boolean;
public
 property BaseType:TBaseType read fbasetype;
 property BaseTypePtr:TTypeInfo read fbasetypeptr write fbasetypeptr;
 property ClassPtr:TClass read fclassptr write fclassptr;
 property ClassRefTypeInfo:TTypeinfo read fclasstypeinfo;
 property FieldCount:Integer read GetFieldCount;
 property Fields[Index:Integer]:TSymbolInfo read GetField write SetField;
 property InstSize:Integer read finstsize write finstsize;
 property InterfaceDecl:Boolean read finterfacedecl write finterfacedecl;
 property Max:Integer read fmax;
 property Method:Boolean read fmethod;
 property MethodType:TMethodType read fmethodtype write fmethodtype;
 property Min:Integer read fmin;
 property Name:String read fname write fname;
 property ResultPtr:Boolean read fresultptr write fresultptr;
 property RTTInfo:PPTypeInfo read frttinfo;
 property Size:Integer read fsize write fsize;
 property UnitInfo:TUnitInfo read funitinfo;
 property VMTSize:Integer read fvmtsize write fvmtsize;
 procedure AddField(si:TSymbolInfo);overload;
 function AddField(const name:string;t:TTypeInfo;vis:TFieldVisibility):TSymbolInfo;overload;
 function AddMethod(const name:string;mt:TMethodType;vis:TFieldVisibility;mp:pointer=nil;f:TSymbolFlags=[];index:integer=0):TSymbolInfo;
 function AddParam(const name:string;t:TTypeInfo;pt:TParamType;flags:TSymbolFlags=[]):TSymbolInfo;
 function FindField(const s:string;context:TTypeInfo;unitinf:TUnitInfo):TSymbolInfo;
 function IsDeclared(const s:string):Boolean;
 function IsFloat:Boolean;
 function IsInteger:Boolean;
 function IsBoolean:Boolean;
 function IsNumeric:Boolean;
 function IsString:Boolean;
 function InheritsFromClass(ti:TTypeInfo):Boolean;
 function CreateOpenArray:TTypeInfo;
 function CreateClassRefType:TTypeInfo;
 function CreateArray(min,max:integer):TTypeInfo;
 procedure MakeVirtual(m:TSymbolInfo);
 procedure OverrideMethod(m:TSymbolInfo);
 function Equals(t2:TTypeInfo):boolean;
 function CanAssignTo(ti:TTypeInfo):boolean;
 procedure CreateClass;
 procedure CreateRTTInfo;
 function CheckMessageFunc:Boolean;
 constructor create(basetype:TBaseType;basetypeptr:TTypeInfo);
 destructor destroy;override;
end;

TCodeGenerator=class
private
 flocalsize,fparamsize,fdeflocalsize,fmaxlocalsize:integer;
 flocalvars:THashedStringList;
 fcode:TMemoryStream;
 lastcmd_lea_edi:integer;
 lastcmd_lea_edi_off:integer;
 lastcmd_add_edi:integer;
 lastcmd_add_edi_off:integer;
 lastcmd_setal_z:integer;
 lastcmd_setal_nz:integer;
 lastcmd_setal_g:integer;
 lastcmd_setal_ge:integer;
 lastcmd_setal_l:integer;
 lastcmd_setal_le:integer;
 cfunc:TSymbolInfo;
 cclass:TTypeInfo;
 fpustacksize:integer;
 fpustacktempvar:TSymbolInfo;
 restoreparamlist:array of TSymbolInfo;
 function GetLocalCount:Integer;
 function GetLocalVar(index:Integer):TSymbolInfo;
 procedure SetLocalVar(index:Integer;s:TSymbolInfo);
 procedure WriteData(const s:string);
public
 edx_count,esi_count,ebx_count:Integer;
 property Code:TMemoryStream read fcode;
 property Context:TTypeInfo read cclass;
 property CurrentFunction:TSymbolInfo read cfunc;
 property LocalVarCount:Integer read GetLocalCount;
 property LocalVar[index:Integer]:TSymbolInfo read GetLocalVar write SetLocalVar;
 procedure AddLocal(si:TSymbolInfo);overload;
 function AddLocal(const name:string;ti:TTypeInfo):TSymbolInfo;overload;
 function FindLocal(const name:string):Integer;
 function GetLocal(const name:string):TSymbolInfo;
 function ConvertDWord(dw:integer):string;
 function FindSymbol(const name:string):TSymbolInfo;
 function FindType(const name:string):TTypeInfo;
 function AllocateTempVar(ti:TTypeInfo):TSymbolInfo;
 procedure CopyRegisterParams;
 procedure CopyRefParams;
 procedure BeginConstructor;
 procedure EndConstructor;
 procedure BeginDestructor;
 procedure EndDestructor;
 procedure BeginFunction(si:TSymbolInfo);
 procedure EndFunction;
 procedure NewStatement;
 procedure CallAddr(p:pointer);
 procedure LoadVar(si:TSymbolInfo);
 procedure LoadVarAddress(si:TSymbolInfo);
 procedure SaveStack;
 procedure RestoreStack;
 procedure IncFPUStack;
 procedure DecFPUStack;
 procedure Mov_EDX_EAX;
 procedure FreeEDX;
 procedure PopFPUReg;
 procedure Mov_ESI_EDI;
 procedure FreeESI;
 procedure Add_EDI_ESI;
 procedure Mov_EDI_EAX;
 procedure Mov_EAX_EDI;
 procedure Movzx_EAX_AL;
 procedure Mov_EBX_EAX;
 procedure Test_EBX;
 procedure PushReg(reg:TRegister);
 procedure PopReg(reg:TRegister);
 procedure PushParamBlock(size:integer);
 procedure PushFloat;
 procedure LoadConst(i:integer);overload;
 procedure LoadConst(f:single);overload;
 procedure LoadConst(t:TTypeInfo);overload;
 procedure LoadConst(const s:string);overload;
 procedure AddEDIAddress(c:Integer);
 procedure AddFieldAddress(si:TSymbolInfo);
 procedure LoadIntVar2Float(si:TSymbolInfo);
 procedure LoadIntRefEDI;
 procedure LoadByteRefEDI;
 procedure LoadFloatRefEDI;
 procedure StoreIntRefEDI;
 procedure StoreIntLocal(si:TSymbolInfo);
 procedure StoreIntLocalConst(si:TSymbolInfo;c:integer);
 procedure StoreByteRefEDI;
 procedure StoreByteLocal(si:TSymbolInfo);
 procedure StoreByteLocalConst(si:TSymbolInfo;b:boolean);
 procedure StoreFloatRefEDI;
 procedure StoreFloatLocal(si:TSymbolInfo);
 procedure StoreFloatLocalConst(si:TSymbolInfo;c:single);
 procedure StoreFloatRefIntEDI;
 procedure IntToFloat;
 procedure MoveBlock(size:integer);
 procedure CallMethod(obj:boolean;si:TSymbolInfo);
 procedure AddIntLocal(si:TSymbolInfo);
 procedure AddIntConst(c:integer);
 procedure AddIntValue;
 procedure AddFloatLocal(si:TSymbolInfo);
 procedure AddFloatConst(s:single);
 procedure AddFloatValue;
 procedure AddStringValue(target:TSymbolInfo);
 procedure IncVar(si:TSymbolInfo);
 procedure SubIntLocal(si:TSymbolInfo);
 procedure SubIntConst(c:integer);
 procedure SubIntValue;
 procedure SubFloatLocal(si:TSymbolInfo);
 procedure SubFloatConst(s:single);
 procedure SubFloatValue;
 procedure DecEBX;
 procedure DecVar(si:TSymbolInfo);
 procedure MultIntLocal(si:TSymbolInfo);
 procedure MultIntConst(c:Integer);
 procedure MultIntValue;
 procedure MultFloatLocal(si:TSymbolInfo);
 procedure MultFloatConst(s:single);
 procedure MultFloatValue;
 procedure DivIntLocal(si:TSymbolInfo);
 procedure DivIntConst(c:Integer);
 procedure DivIntValue;
 procedure DivFloatLocal(si:TSymbolInfo);
 procedure DivFloatConst(s:single);
 procedure DivFloatValue;
 procedure ModIntLocal(si:TSymbolInfo);
 procedure ModIntConst(c:Integer);
 procedure ModIntValue;
 procedure EqualIntLocal(si:TSymbolInfo);
 procedure EqualIntConst(c:Integer);
 procedure EqualIntValue;
 procedure EqualFloatLocal(si:TSymbolInfo);
 procedure EqualFloatConst(s:single);
 procedure EqualFloatValue;
 procedure EqualStringValue;
 procedure NotEqualIntLocal(si:TSymbolInfo);
 procedure NotEqualIntConst(c:Integer);
 procedure NotEqualIntValue;
 procedure NotEqualFloatLocal(si:TSymbolInfo);
 procedure NotEqualFloatConst(s:single);
 procedure NotEqualFloatValue;
 procedure NotEqualStringValue;
 procedure NotBoolean;
 procedure NegInteger;
 procedure NegFloat;
 procedure GreaterIntLocal(si:TSymbolInfo);
 procedure GreaterIntConst(c:Integer);
 procedure GreaterIntValue;
 procedure GreaterFloatLocal(si:TSymbolInfo);
 procedure GreaterFloatConst(s:single);
 procedure GreaterFloatValue;
 procedure GreaterEqualIntLocal(si:TSymbolInfo);
 procedure GreaterEqualIntConst(c:Integer);
 procedure GreaterEqualIntValue;
 procedure GreaterEqualFloatLocal(si:TSymbolInfo);
 procedure GreaterEqualFloatConst(s:single);
 procedure GreaterEqualFloatValue;
 procedure LessIntLocal(si:TSymbolInfo);
 procedure LessIntConst(c:Integer);
 procedure LessIntValue;
 procedure LessFloatLocal(si:TSymbolInfo);
 procedure LessFloatConst(s:single);
 procedure LessFloatValue;
 procedure LessEqualIntLocal(si:TSymbolInfo);
 procedure LessEqualIntConst(c:Integer);
 procedure LessEqualIntValue;
 procedure LessEqualFloatLocal(si:TSymbolInfo);
 procedure LessEqualFloatConst(s:single);
 procedure LessEqualFloatValue;
 procedure AndIntLocal(si:TSymbolInfo);
 procedure AndBooleanLocal(si:TSymbolInfo);
 procedure AndConst(c:integer);
 procedure AndValue;
 procedure OrIntLocal(si:TSymbolInfo);
 procedure OrBooleanLocal(si:TSymbolInfo);
 procedure OrConst(c:integer);
 procedure OrValue;
 procedure XorIntLocal(si:TSymbolInfo);
 procedure XorBooleanLocal(si:TSymbolInfo);
 procedure XorConst(c:Integer);
 procedure XorValue;
 function JumpTrue:Integer;
 function JumpFalse:Integer;
 function Jump_Zero:Integer;
 function Jump:Integer;
 procedure SetJumpTarget(jumpop,target:Integer);
 constructor create;
 destructor destroy;override;
end;

TUnitInfo=class
private
 ftypelist:THashedStringList;
 fsymbollist:THashedStringList;
 interfaceimplist:TList;
 fname:string;
 finterfacesec:boolean;
 stringconstlist:TList;
public
 property InterfaceSec:Boolean read finterfacesec write finterfacesec;
 property Name:string read fname;
 procedure AddImport(unitinfo:TUnitInfo);
 function AddClass(const cname:string;baseclass:TTypeInfo):TTypeInfo;
 function AddRecord(const cname:string;basetype:TTypeInfo):TTypeInfo;
 function AddType(const tname:string;basetype:TBaseType):TTypeInfo;
 function AddEnum(const tname:string;const enumlist:array of string):TTypeInfo;
 function AddSymbol(const name:string;t:TTypeInfo):TSymbolInfo;
 function AddConst(const name,val:string;t:TTypeInfo):TSymbolInfo;
 function AddFunction(const name:string;mt:TMethodType):TSymbolInfo;
 function AddStringConst(const name:string):Integer;
 function FindType(const name:string;noimp:boolean=false):TTypeInfo;
 function FindSymbol(const name:string;noimp:boolean=false):TSymbolInfo;
 function IsDeclared(const name:string):boolean;
 function CheckImplemention:TSymbolInfo;
 class function FindUnit(const name:string):TUnitInfo;
 procedure RegisterClasses;
 procedure UnRegisterClasses;
 constructor create(const name:string);
 destructor destroy;override;
end;

TTypeInfoClass=class(TPersistent)
private
 fint:Integer;
 fsingle:single;
 fboolean:boolean;
 fstring:shortstring;
published
 property IntProp:Integer read fint;
 property SingleProp:Single read fsingle;
 property BooleanProp:boolean read fboolean;
 property StrProp:ShortString read fstring;
end;

var
 System_Integer:TTypeInfo;
 System_Untyped:TTypeInfo;
 System_Single:TTypeInfo;
 System_String:TTypeInfo;
 System_Nil:TTypeInfo;
 System_Procedure:TTypeInfo;
 System_Boolean:TTypeInfo;
 System_StringCopy:Pointer;
 System_StringConcat:Pointer;
 System_StringEqual:Pointer; 

procedure InitSystemUnit;

implementation

var
 unitlist:TObjectList;
 _classcreate:pointer;
 _afterconstruction:pointer;
 _beforedestruction:pointer;
 _classdestroy:pointer;

function Align(offset:Integer):Integer;
begin
if offset mod 4=0 then result:=offset else result:=offset+4-(offset mod 4);
end;

// TSymbolInfo

constructor TSymbolInfo.create(const n:string;t:TTypeInfo);
begin
 inherited create;
 fname:=n;
 ftypeinfo:=t;
end;

destructor TSymbolInfo.destroy;
begin
if fexecutemem<>nil then
begin
 VirtualFree(fexecutemem,0,MEM_RELEASE);
 fexecutemem:=nil;
end;
inherited;
end;

// TTypeInfo

function TTypeInfo.GetFieldCount:Integer;
begin
 result:=ffields.count;
end;

function TTypeInfo.GetField(i:integer):TSymbolInfo;
begin
 result:=TSymbolInfo(ffields.objects[i]);
end;

procedure TTypeInfo.SetField(i:integer;v:TSymbolInfo);
begin
 ffields.objects[i]:=v;
 ffields[i]:=v.Name;
end;

procedure TTypeInfo.AddField(si:TSymbolInfo);
begin
 si.Offset:=finstsize;
 si.fparent:=self;
 ffields.AddObject(si.Name,si);
 finstsize:=finstsize+si.TypeInfo.Size;
 if BaseType<>btclass then fsize:=fsize+si.TypeInfo.size;
end;

function TTypeInfo.AddField(const name:string;t:TTypeInfo;vis:TFieldVisibility):TSymbolInfo;
begin
if IsDeclared(name) then raise ESyntaxError.Create('Identifier redeclared "'+name+'"');
result:=TSymbolInfo.create(name,t);
result.funitinfo:=UnitInfo;
result.fVisibility:=vis;
AddField(result);
end;

function TTypeInfo.AddMethod(const name:string;mt:TMethodType;vis:TFieldVisibility;mp:Pointer=nil;f:TSymbolFlags=[];index:integer=0):TSymbolInfo;
var
 m:TTypeInfo;
begin
if IsDeclared(name) then raise ESyntaxError.Create('Identifier redeclared "'+name+'"');
m:=UnitInfo.AddType(self.name+'.'+name,btmethod);
m.MethodType:=mt;
result:=TSymbolInfo.create(name,m);
result.fUnitInfo:=UnitInfo;
result.fVisibility:=vis;
result.fmethodptr:=mp;
result.TypeInfo.fmethod:=true;
if mt = mtconstructor then
 result.TypeInfo.BaseTypePtr:=self;

AddField(result);
if mp<>nil then
begin
 if mt in [mtclassprocedure,mtclassfunction] then
  result.TypeInfo.AddParam('Self',self.ClassRefTypeInfo,ptvalue)
 else
  result.TypeInfo.AddParam('Self',self,ptvalue);
 if mt in [mtconstructor,mtdestructor] then
  result.TypeInfo.AddParam('',System_Integer,ptvalue);
end;
result.fflags:=result.fflags+f;
result.fvmtindex:=index;
if sfvirtual in f then MakeVirtual(result);
end;

function TTypeInfo.AddParam(const name:string;t:TTypeInfo;pt:TParamType;flags:TSymbolFlags=[]):TSymbolInfo;
begin
if IsDeclared(name) then raise ESyntaxError.Create('Identifier redeclared "'+name+'"');
result:=TSymbolInfo.create(name,t);
result.fflags:=flags;
result.fparamtype:=pt;
if ((result.TypeInfo.size>4)) then
 include(result.fFlags,sfreference);
//if (pt=ptvar) or  ((result.TypeInfo.size>4) and (pt=ptconst)) then
// include(result.fFlags,sfreference);
result.fUnitInfo:=UnitInfo;
if (sfreference in result.Flags) or (t.BaseType in [btinteger,btboolean,btenum,btclass,btclassref]) then
begin
 if integer(fusedparamreg)<integer(regecx) then
 begin
  inc(fusedparamreg);
  result.fregister:=fusedparamreg;
 end;
 if (fusedparamreg=regeax) and (MethodType in [mtconstructor,mtdestructor]) then
  inc(fusedparamreg); // edx ist beim constructor/destructor reserviert
end;
AddField(result);
end;

function TTypeInfo.FindField(const s:string;context:TTypeInfo;unitinf:TUnitInfo):TSymbolInfo;
var
 i:integer;
begin
if BaseType=btclassref then
begin
 result:=BaseTypePtr.FindField(s,context,unitinf);
 if result.ftypeinfo.BaseType<>btmethod then result:=nil;
 if not (result.ftypeinfo.MethodType in [mtconstructor,mtclassprocedure,mtclassfunction]) then result:=nil;
 exit;
end;
i:=ffields.IndexOf(s);
if i=-1 then result:=nil else result:=TSymbolInfo(ffields.objects[i]);

if (result<>nil) and (unitinf<>result.UnitInfo) then
begin
 if (result.Visibility=fsprivate) or (context=nil) or not context.InheritsFromClass(self) then result:=nil;
end;

if (result=nil) and (BaseTypePtr<>nil) then result:=BaseTypePtr.FindField(s,context,unitinf);
end;

function TTypeInfo.IsDeclared(const s:string):Boolean;
begin
result:=ffields.IndexOf(s)<>-1;
end;

function TTypeInfo.IsFloat:Boolean;
begin
result:=BaseType=btsingle;
end;

function TTypeInfo.IsInteger:Boolean;
begin
result:=BaseType=btinteger;
end;

function TTypeInfo.IsBoolean:Boolean;
begin
result:=self=System_Boolean;
end;

function TTypeInfo.IsNumeric:Boolean;
begin
result:=BaseType in [btinteger,btsingle];
end;

function TTypeInfo.IsString:Boolean;
begin
result:=BaseType=btstring;
end;

function TTypeInfo.InheritsFromClass(ti:TTypeInfo):Boolean;
begin
result:=(self=ti) or (assigned(BaseTypePtr) and BaseTypePtr.InheritsFromClass(ti));
end;

function TTypeInfo.CreateOpenArray:TTypeInfo;
begin
result:=unitinfo.AddType('array of '+Name,btopenarray);
result.fbasetypeptr:=self;
end;

function TTypeInfo.CreateArray(min,max:integer):TTypeInfo;
begin
result:=unitinfo.AddType('array['+inttostr(min)+'..'+inttostr(max)+'] of '+Name,btarray);
result.fmin:=min;
result.fmax:=max;
result.fSize:=(max-min+1)*Size;
result.fbasetypeptr:=self;
end;

function TTypeInfo.CreateClassRefType:TTypeInfo;
begin
result:=unitinfo.AddType('class of '+Name,btclassref);
result.fSize:=4;
result.finstsize:=4;
result.fBaseTypePtr:=self;
end;

function TTypeInfo.CheckMessageFunc:Boolean;
begin
result:=true;
if MethodType<>mtprocedure then result:=false;
if FieldCount<>2 then
 result:=false
else
begin
 if (Fields[1].TypeInfo.BaseType<>btrecord) or (Fields[1].ParamType<>ptvar) then
  result:=false;
end;  
end;

procedure TTypeInfo.MakeVirtual(m:TSymbolInfo);
begin
m.fvmtindex:=vmtsize;inc(fvmtsize);
include(m.fflags,sfvirtual);
end;

procedure TTypeInfo.OverrideMethod(m:TSymbolInfo);
var
 pmethod:TSymbolInfo;
 i:integer;
begin
if BaseTypePtr<>nil then pmethod:=BaseTypePtr.FindField(m.Name,self,UnitInfo) else pmethod:=nil;
if pmethod=nil then raise ESyntaxError.Create('Method "'+m.Name+'" does not exist in base class.');
//if not pmethod.ftypeinfo.Equals(m.TypeInfo) then
if not m.TypeInfo.Equals(pmethod.TypeInfo) then
 raise ESyntaxError.Create('Declaration of "'+m.Name+'" differs from previous declaration.');
for i:=0 to m.TypeInfo.FieldCount-1 do
begin
 if not SameText(m.TypeInfo.Fields[i].Name,pmethod.TypeInfo.Fields[i].Name) then
  raise ESyntaxError.Create('Declaration of "'+m.Name+'" differs from previous declaration.');
end;
include(m.fflags,sfvirtual);
m.fvmtindex:=pmethod.VMTIndex;

end;

function TTypeInfo.MethodEquals(t2:TTypeInfo):boolean;
var
 i:integer;
begin
result:=false;
if not t2.BaseTypePtr.Equals(BaseTypePtr) then exit;
if ffields.count<>t2.ffields.count then exit;
if MethodType<>t2.MethodType then exit;
for i:=0 to FieldCount-1 do
begin
 if not (Fields[i].TypeInfo.Equals(t2.Fields[i].TypeInfo) or
 Fields[i].TypeInfo.InheritsFromClass(t2.Fields[i].TypeInfo)) then exit;
 if Fields[i].ParamType<>t2.Fields[i].ParamType then exit;
end;
result:=true;
end;

function TTypeInfo.Equals(t2:TTypeInfo):boolean;
begin
if (t2=self) then
begin
 result:=true;
 exit;
end;
if (t2.BaseType<>BaseType) then
begin
 result:=false;
 exit;
end;
case BaseType of
 btinteger:result:=true;
 btboolean:result:=true;
 btenum:result:=false;
 btsingle:result:=true;
 btstring:result:=true;
 btrecord:result:=false;
 btarray:result:=BaseTypePtr.Equals(t2.BaseTypePtr);
 btopenarray:result:=BaseTypePtr.Equals(t2.BaseTypePtr);
 btclass:result:=false;
 btclassref:result:=BaseTypePtr.Equals(t2.BaseTypePtr);
 btmethod:result:=MethodEquals(t2);
 else result:=false;
end; 
end;

function TTypeInfo.CanAssignTo(ti:TTypeInfo):boolean;
begin
result:=Equals(ti);
result:=result or ((BaseType=btinteger) and (ti.BaseType=btsingle));
result:=result or ((BaseType=btnil) and (ti.BaseType in [btclass,btclassref]));
if (BaseType=btclass) and  (ti.BaseType=btclass) then
 result:=result or InheritsFromClass(ti);
if (BaseType=btclassref) and  (ti.BaseType=btclassref) then
 result:=result or BaseTypePtr.InheritsFromClass(ti.BaseTypePtr);
end;

procedure TTypeInfo.CreateClass;
var
 i,numdyn:integer;
 p:pointer;
begin
numdyn:=0;
for i:=0 to FieldCount-1 do
 if sfmessage in Fields[i].Flags then inc(numdyn);
fshortname:=UnitInfo.Name+'.'+name;
GetMem(fvmtptr,76+vmtsize*4);
fClassPtr:=pointer(integer(fvmtptr)+76);
move(pointer(integer(BaseTypePtr.ClassPtr)-76)^,fvmtptr^,76+BaseTypePtr.vmtsize*4);
PPointer(integer(fClassPtr)+system.vmtClassName)^:=@fshortname;
PInteger(integer(fClassPtr)+system.vmtInstanceSize)^:=InstSize;
PPointer(integer(fClassPtr)+system.vmtSelfPtr)^:=fclassptr;
fparentvmt:=BaseTypePtr.ClassPtr;
PPointer(integer(fClassPtr)+system.vmtParent)^:=@fparentvmt;

for i:=0 to FieldCount-1 do
 if (fields[i].TypeInfo.BaseType=btmethod) and (sfvirtual in fields[i].Flags) then
 begin
  PPointer(integer(fClassPtr)+fields[i].fvmtindex*4)^:=Fields[i].MethodPtr;
 end;

getmem(fdyntabptr,2+numdyn*6);
PPointer(integer(fClassPtr)+system.vmtDynamicTable)^:=fdyntabptr;
p:=fdyntabptr;
PWord(p)^:=numdyn;
inc(PWord(p));
for i:=0 to FieldCount-1 do
 if sfmessage in Fields[i].Flags then
 begin
  PWord(p)^:=Fields[i].MessageID;
  inc(PWord(p));
 end;
for i:=0 to FieldCount-1 do
 if sfmessage in Fields[i].Flags then
 begin
  PPointer(p)^:=Fields[i].MethodPtr;
  inc(PPointer(p));
 end;

unregister:=true;
end;

procedure TTypeInfo.CreateRTTInfo;
const
 PublishedTypes:set of TBaseType=[btinteger,btenum,btboolean,btsingle,btstring,btclass];
var
 i,size,numprop:integer;
 namelist:string;
 p:pointer;
begin
frttinfo:=@frttdata;
size:=sizeof(TTypeKind)+1+length(name);
case BaseType of
 btclass:
 begin
  size:=size+4+4+2+1+length(UnitInfo.Name);
  numprop:=0;
  for i:=0 to FieldCount-1 do
   if (Fields[i].Visibility=fspublished) and (Fields[i].TypeInfo.BaseType in PublishedTypes) then
   begin
   inc(size,27+length(fields[i].Name));
   inc(numprop);
   end;
  size:=size+4;
  getmem(frttdata,size);
  p:=frttdata;
  TTypeKind(p^):=typinfo.tkclass;inc(PByte(p),sizeof(TTypeKind));
  PShortString(p)^:=Name;inc(PByte(p),length(name)+1);
  PPointer(p)^:=fclassptr;inc(PByte(p),4);
  PPointer(integer(frttdata)+size-4)^:=ClassPtr.ClassParent.ClassInfo;
  PPointer(p)^:=pointer(integer(frttdata)+size-4);inc(PByte(p),4);
  PSmallInt(p)^:=numprop;inc(PByte(p),2);
  PShortString(p)^:=UnitInfo.Name;inc(PByte(p),length(UnitInfo.Name)+1);
  PSmallInt(p)^:=numprop;inc(PByte(p),2);  
  for i:=0 to FieldCount-1 do
   if (Fields[i].Visibility=fspublished) and (Fields[i].TypeInfo.BaseType in PublishedTypes) then
   begin
    with PPropInfo(p)^ do
    begin
     PropType:=Fields[i].TypeInfo.RTTInfo;
     GetProc:=pointer(Cardinal(Fields[i].Offset) or $FF000000);
     SetProc:=pointer(Cardinal(Fields[i].Offset) or $FF000000);
     StoredProc:=nil;
     Index:=0;
     Default:=0;
     NameIndex:=i;
     Name:=Fields[i].Name;
    end;
    inc(PByte(p),27+length(fields[i].Name));
   end;
    PPointer(integer(fClassPtr)+system.vmtTypeInfo)^:=frttdata;   
 end;
 btenum:
 begin
  exit;
  for i:=0 to FieldCount-1 do
   namelist:=namelist+chr(length(Fields[i].Name))+Fields[i].Name;
  getmem(frttdata,size+sizeof(TOrdType)+4+4+length(namelist));
  frttdata.Kind:=typinfo.tkEnumeration;
  frttdata.Name:=name;
  GetTypeData(frttdata)^.OrdType:=otSLong;
  GetTypeData(frttdata)^.MinValue:=0;
  GetTypeData(frttdata)^.MaxValue:=FieldCount-1;
  GetTypeData(frttdata)^.BaseType:=System_Integer.RTTInfo;
  GetTypeData(frttdata)^.NameList:=namelist;
 end;
end;
end;

constructor TTypeInfo.create(basetype:TBaseType;basetypeptr:TTypeInfo);
begin
 inherited create;
 fbasetypeptr:=basetypeptr;
 fbasetype:=basetype;
 ffields:=THashedStringList.Create;
 ffields.CaseSensitive:=false;
 if basetypeptr<>nil then
 begin
  fsize:=basetypeptr.fsize;
  finstsize:=basetypeptr.finstsize;
  vmtsize:=basetypeptr.vmtsize;
 end; 
end;

destructor TTypeInfo.destroy;
var
 i:integer;
begin
 FreeMem(frttdata);
 freemem(fvmtptr);
 freemem(fdyntabptr);
 for i:=0 to FieldCount-1 do
  Fields[i].Free;
 ffields.free;
 inherited;
end;

// TCodeGenerator

function TCodeGenerator.GetLocalCount:Integer;
begin
 result:=flocalvars.Count;
end;

function TCodeGenerator.GetLocalVar(index:Integer):TSymbolInfo;
begin
 result:=TSymbolInfo(flocalvars.objects[index]);
end;

procedure TCodeGenerator.SetLocalVar(index:Integer;s:TSymbolInfo);
begin
 flocalvars.objects[index]:=s;
end;

procedure TCodeGenerator.WriteData(const s:string);
begin
code.Write(s[1],length(s));
end;

procedure TCodeGenerator.AddLocal(si:TSymbolInfo);
var
 l:integer;
begin
if (sfreference in si.Flags) and (si.ParamType=ptvalue) then
begin
 si.RefParam:=TSymbolInfo.create('',si.TypeInfo);
 si.RefParam.fregister:=si.ParamRegister;
 include(si.RefParam.fflags,sfreference);
 si.ParamType:=ptvar;
 AddLocal(si.RefParam);
 si.ParamType:=ptnone;
 exclude(si.fflags,sfreference);
 AddLocal(si);
 l:=length(restoreparamlist);
 setlength(restoreparamlist,l+1);
 restoreparamlist[l]:=si;
 //include(si.fflags,sfreference);
 //si.ParamType:=ptvalue;
 exit;
end;
if (si.ParamType<>ptnone) and (si.ParamRegister=regnone) then
begin
 si.Offset:=8+fparamsize;
 if sfreference in si.Flags then
  fparamsize:=fparamsize+4
 else
  fparamsize:=fparamsize+si.TypeInfo.Size;
 flocalvars.AddObject(si.Name,si);
end
else
begin
 flocalvars.AddObject(si.Name,si);
 if sfreference in si.Flags then
  flocalsize:=flocalsize+4
 else
  flocalsize:=align(flocalsize+si.TypeInfo.Size);
// flocalsize:=align(flocalsize+si.TypeInfo.Size);
 si.Offset:=-flocalsize;
end;
end;

function TCodeGenerator.AddLocal(const name:string;ti:TTypeInfo):TSymbolInfo;
var
 si:TSymbolInfo;
begin
si:=TSymbolInfo.create(name,ti);
AddLocal(si);
result:=si;
end;

function TCodeGenerator.FindLocal(const name:string):Integer;
begin
result:=flocalvars.IndexOf(name);
end;

function TCodeGenerator.GetLocal(const name:string):TSymbolInfo;
var
 i:integer;
begin
i:=FindLocal(name);
if i<>-1 then result:=LocalVar[i] else result:=nil;
end;

function TCodeGenerator.ConvertDWord(dw:integer):string;
begin
 result:=chr(dw and $FF)+chr((dw and $FF00) shr 8)+chr((dw and $FF0000) shr 16)+chr((dw and $FF000000) shr 24);
end;

function TCodeGenerator.FindSymbol(const name:string):TSymbolInfo;
begin
result:=GetLocal(name);
//if (result=nil) and (cfunc.parent<>nil) then
// result:=cfunc.Parent.FindField(name,cfunc.Parent);
if result=nil then
 result:=cfunc.UnitInfo.FindSymbol(name);

end;

function TCodeGenerator.FindType(const name:string):TTypeInfo;
begin
result:=cfunc.UnitInfo.FindType(name);
end;

function TCodeGenerator.AllocateTempVar(ti:TTypeInfo):TSymbolInfo;
begin
result:=AddLocal('',ti);
end;

procedure TCodeGenerator.CopyRegisterParams;
var
 i:integer;
begin
for i:=0 to LocalVarCount-1 do
 if LocalVar[i].RefParam=nil then
 case LocalVar[i].ParamRegister of
  regeax:WriteData(#$89#$85+ConvertDWord(LocalVar[i].Offset));
  regecx:WriteData(#$89#$8D+ConvertDWord(LocalVar[i].Offset));
  regedx:WriteData(#$89#$95+ConvertDWord(LocalVar[i].Offset));
 end; 
end;

procedure TCodeGenerator.CopyRefParams;
var
 i:integer;
begin
for i:=0 to LocalVarCount-1 do
 if LocalVar[i].RefParam<>nil then
 begin
  LoadVarAddress(LocalVar[i].RefParam);
  Mov_ESI_EDI;
  LoadVarAddress(LocalVar[i]);
  MoveBlock(LocalVar[i].TypeInfo.Size);
 end;
end;

procedure TCodeGenerator.BeginConstructor;
begin
WriteData(#$84#$D2); // TEST dl,dl
WriteData(#$74#$08); // jz+$08
WriteData(#$83#$C4#$F0); // ADD esp,-$10
CallAddr(@_ClassCreate);
WriteData(#$52);         // PUSH edx
WriteData(#$50);         // PUSH eax
end;

procedure TCodeGenerator.EndConstructor;
begin
WriteData(#$58);     // POP eax
WriteData(#$5A);     // POP edx
WriteData(#$84#$D2); // TEST dl,dl
WriteData(#$74#$0F); // jz +$0f
CallAddr(@_AfterConstruction);
WriteData(#$64#$8F#$05#$00#$00#$00#$00); // pop dword ptr fs:[0]
WriteData(#$83#$C4#$0C); // add esp,$0C
end;

procedure TCodeGenerator.BeginDestructor;
begin
CallAddr(@_BeforeDestruction);
WriteData(#$52);         // PUSH edx
WriteData(#$50);         // PUSH eax

end;

procedure TCodeGenerator.EndDestructor;
begin
WriteData(#$58);     // POP eax
WriteData(#$5A);     // POP edx
WriteData(#$84#$D2); // TEST dl,dl
WriteData(#$7E#$07); // jle +$07
CallAddr(@_classdestroy);
end;


procedure TCodeGenerator.BeginFunction(si:TSymbolInfo);
var
 i:integer;
begin
FreeAndNil(fcode);
fcode:=TMemoryStream.Create;
flocalsize:=0;
fparamsize:=0;

for i:=0 to flocalvars.count-1 do
 if TSymbolInfo(flocalvars.objects[i]).ParamType=ptnone then
  TSymbolInfo(flocalvars.objects[i]).free;
flocalvars.Clear;
cfunc:=si;
cclass:=cfunc.Parent;

for i:=cfunc.TypeInfo.FieldCount-1 downto 0 do
 AddLocal(cfunc.TypeInfo.Fields[i]);

WriteData(#$55#$8B#$EC); // PUSH ebp; mov ebp,esp
WriteData(#$81#$c4+ConvertDWord(-flocalsize)); // add esp,0
case cfunc.TypeInfo.MethodType of
 mtconstructor:BeginConstructor;
 mtdestructor:BeginDestructor;
end;

CopyRegisterParams;
CopyRefParams;

if (si.TypeInfo.BaseTypePtr<>nil) and (cfunc.TypeInfo.FindField('Result',si.Parent,si.UnitInfo)=nil) then
 AddLocal('Result',si.TypeInfo.BaseTypePtr);


WriteData(#$57);         // PUSH edi
WriteData(#$56);         // PUSH esi
PushReg(regebx);
fmaxlocalsize:=flocalsize;
fdeflocalsize:=flocalsize;
NewStatement;
end;

procedure TCodeGenerator.EndFunction;
var
 i:integer;
begin
flocalsize:=fmaxlocalsize;
case cfunc.TypeInfo.MethodType of
 mtfunction:
 begin
  if cfunc.TypeInfo.BaseTypePtr.BaseType in [btinteger,btboolean,btenum,btclass,btclassref] then
  begin
   LoadVar(FindSymbol('Result'));
  // mov eax,result
  end
  else
  if cfunc.TypeInfo.BaseTypePtr.BaseType=btsingle then
  begin
   LoadVar(FindSymbol('Result'));
  // fld result
  end;
 end;
end;

PopReg(regebx);
WriteData(#$5E);         // POP esi
WriteData(#$5F);         // POP edi

case cfunc.TypeInfo.MethodType of
 mtconstructor:EndConstructor;
 mtdestructor:EndDestructor;
end;

WriteData(#$81#$c4+ConvertDWord(flocalsize)); // add esp,flocalsize
WriteData(#$5D); // pop ebp
WriteData(#$C2+chr(fparamsize and $FF)+chr(fparamsize shr 8 and $FF)); // ret paramsize

code.Seek(0,soFromBeginning);
WriteData(#$55#$8B#$EC); // PUSH ebp; mov ebp,esp
WriteData(#$81#$c4+ConvertDWord(-flocalsize));

cfunc.fexecutemem:=VirtualAlloc(nil,code.Size,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
move(code.memory^,cfunc.fexecutemem^,code.size);
cfunc.fmethodptr:=cfunc.fexecutemem;
fcode.SetSize(0);
NewStatement;
for i:=0 to length(restoreparamlist)-1 do
 with restoreparamlist[i] do
 begin
 include(fflags,sfreference);
 ParamType:=ptvalue;
 end;
restoreparamlist:=nil;
end;

procedure TCodeGenerator.NewStatement;
begin
lastcmd_lea_edi:=-1;
lastcmd_add_edi:=-1;
lastcmd_setal_z:=-1;
lastcmd_setal_nz:=-1;
lastcmd_setal_g:=-1;
lastcmd_setal_ge:=-1;
lastcmd_setal_l:=-1;
lastcmd_setal_le:=-1;

fpustacksize:=0;
if flocalsize>fmaxlocalsize then fmaxlocalsize:=flocalsize;
flocalsize:=fdeflocalsize;
edx_count:=0;
esi_count:=0;
end;

procedure TCodeGenerator.CallAddr(p:pointer);
begin
WriteData(#$FF#$15+ConvertDWord(integer(p))); // CALL [p]
end;

procedure TCodeGenerator.LoadVar(si:TSymbolInfo);
begin
case si.TypeInfo.BaseType of
 btinteger,btenum,btclass,btclassref:
 begin
  WriteData(#$8B#$85+ConvertDWord(si.Offset)); // mov eax,[si.offset]
  if sfreference in si.Flags then WriteData(#$8B#$00) // mov eax,[eax]
 end;
 btboolean:
 begin
  WriteData(#$8B#$85+ConvertDWord(si.Offset)); // mov eax,[si.offset]
  if sfreference in si.Flags then WriteData(#$0F#$B6#$00) // movzx eax,[eax]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$85+ConvertDWord(si.Offset)); // mov eax,[si.offset]
   WriteData(#$D9#$00); // fld [eax]
   IncFPUStack;
  end
  else
  begin
  WriteData(#$D9#$85+ConvertDWord(si.Offset)); // fld [si.offset]
   IncFPUStack;
  end;
 end;

 btrecord,btarray,btstring:LoadVarAddress(si);
end;
end;

procedure TCodeGenerator.LoadVarAddress(si:TSymbolInfo);
begin
if sfreference in si.Flags then
 WriteData(#$8B#$BD+ConvertDWord(si.Offset)) // mov edi,[si.offset]
else
begin
 WriteData(#$8D#$BD+ConvertDWord(si.Offset)); // lea edi,[si.offset]
 lastcmd_lea_edi:=code.Size;
 lastcmd_lea_edi_off:=si.offset;
end;
end;

procedure TCodeGenerator.SaveStack;
var
 i:integer;
 ti:TTypeInfo;
begin
if fpustacksize=0 then exit;
ti:=System_Single.CreateArray(0,fpustacksize-1);
fpustacktempvar:=AllocateTempVar(ti);
for i:=0 to fpustacksize-1 do
 WriteData(#$D9#$9D+ConvertDWord(fpustacktempvar.Offset+i*4));
end;

procedure TCodeGenerator.RestoreStack;
var
 i:integer;
begin
if fpustacksize=0 then exit;
WriteData(#$D9#$5C#$24#$FC); // fstp [esp-4]
for i:=fpustacksize-1 downto 0 do
 WriteData(#$D9#$85+ConvertDWord(fpustacktempvar.Offset+i*4)); // FLD [ebp-4-4*i]
WriteData(#$D9#$44#$24#$FC); // fld [esp-4]
end;

procedure TCodeGenerator.IncFPUStack;
begin
inc(fpustacksize);
end;

procedure TCodeGenerator.DecFPUStack;
begin
dec(fpustacksize);
end;


procedure TCodeGenerator.Mov_EDX_EAX;
begin
if edx_count>0 then
 PushReg(regedx);
WriteData(#$89#$C2); // mov edx,eax
inc(edx_count);
end;

procedure TCodeGenerator.FreeEDX;
begin
dec(edx_count);
if edx_count>0 then PopReg(regedx);
end;

procedure TCodeGenerator.PopFPUReg;
begin
WriteData(#$D9#$5C#$24#$FC);         // fstp [esp-4]
DecFPUStack;
end;

procedure TCodeGenerator.Mov_ESI_EDI;
begin
if esi_count>0 then
 PushReg(regesi);
WriteData(#$89#$FE); // mov esi,edi
inc(esi_count);
end;

procedure TCodeGenerator.FreeESI;
begin
dec(esi_count);
if esi_count>0 then PopReg(regesi);
end;

procedure TCodeGenerator.Add_EDI_ESI;
begin
WriteData(#$01#$F7); // mov esi,edi
FreeESI;
end;

procedure TCodeGenerator.Mov_EDI_EAX;
begin
WriteData(#$89#$C7); // mov edi,eax
end;

procedure TCodeGenerator.Mov_EAX_EDI;
begin
WriteData(#$89#$F8); // mov eax,edi
end;

procedure TCodeGenerator.Movzx_EAX_AL;
begin
WriteData(#$0F#$B6#$C0); // movzx eax,al
end;

procedure TCodeGenerator.Mov_EBX_EAX;
begin
if ebx_count>0 then PushReg(regebx);
inc(ebx_count);
WriteData(#$89#$C3); // mov ebx,eax
end;

procedure TCodeGenerator.Test_EBX;
begin
WriteData(#$85#$DB); // test ebx,ebx
end;

procedure TCodeGenerator.PushReg(reg:TRegister);
begin
case reg of
 regeax:WriteData(#$50);         // PUSH eax
 regedx:WriteData(#$52);         // PUSH edx
 regecx:WriteData(#$51);         // PUSH ecx
 regebx:WriteData(#$53);         // PUSH ebx
 regedi:WriteData(#$57);         // PUSH edi
 regesi:WriteData(#$56);         // PUSH esi
end;
end;

procedure TCodeGenerator.PopReg(reg:TRegister);
begin
case reg of
 regeax:WriteData(#$58);         // POP eax
 regedx:WriteData(#$5A);         // POP edx
 regecx:WriteData(#$59);         // POP ecx
 regebx:WriteData(#$5B);         // POP ebx
 regedi:WriteData(#$5F);         // POP edi
 regesi:WriteData(#$5E);         // POP esi
end;
end;

procedure TCodeGenerator.PushParamBlock(size:integer);
begin
WriteData(#$B9+ConvertDWord(size div 4)); // push [edi]
WriteData(#$FF#$37); // push [edi]
WriteData(#$83#$C7#$04); // add edi,4
WriteData(#$83#$C1#$FF); // add ecx,-1
WriteData(#$75#$F6); // jnz loop

end;

procedure TCodeGenerator.PushFloat;
begin
WriteData(#$83#$C4#$FC); // add esp,-4
WriteData(#$D9#$1C#$24); // fstp [esp]
DecFPUStack;
end;

procedure TCodeGenerator.LoadConst(i:integer);
begin
WriteData(#$B8+ConvertDWord(i));
end;

procedure TCodeGenerator.LoadConst(f:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@f)^)); // mov [esp-4],f
WriteData(#$D9#$44#$24#$FC); // fld [esp-4]
IncFPUStack;
end;

procedure TCodeGenerator.LoadConst(t:TTypeInfo);
begin
WriteData(#$8B#$05+ConvertDWord(integer(@t.ClassPtr)));
end;

procedure TCodeGenerator.LoadConst(const s:string);
begin
WriteData(#$BF+ConvertDWord(cfunc.UnitInfo.AddStringConst(s)));
end;

procedure TCodeGenerator.AddEDIAddress(c:Integer);
begin
//lastcmd_lea_edi:=-1;
//lastcmd_add_edi:=-1;
if lastcmd_lea_edi=code.Size then
begin
 Code.Seek(-6,soFromCurrent);
 WriteData(#$8D#$BD+ConvertDWord(c+lastcmd_lea_edi_off)); // lea edi,[si.offset]
 lastcmd_lea_edi_off:=c+lastcmd_lea_edi_off;
end
else
begin
 if lastcmd_add_edi=code.Size then
 begin
  Code.Seek(-6,soFromCurrent);
  lastcmd_add_edi_off:=lastcmd_add_edi_off+c;
  WriteData(#$81#$C7+ConvertDWord(lastcmd_add_edi_off)); // add edi, si.value
 end
 else
 begin
  if c<>0 then WriteData(#$81#$C7+ConvertDWord(c)); // add edi, si.value
  lastcmd_add_edi:=Code.Size;
  lastcmd_add_edi_off:=c;
 end;
end;
end;

procedure TCodeGenerator.AddFieldAddress(si:TSymbolInfo);
begin
if lastcmd_lea_edi=code.Size then
begin
Code.Seek(-6,soFromCurrent);
if si.Parent.BaseType=btclass then
begin
 WriteData(#$8B#$BD+ConvertDWord(lastcmd_lea_edi_off)); // mov edi,[si.offset]
 if si.Offset<>0 then WriteData(#$81#$C7+ConvertDWord(si.Offset));  // add edi, si.value
 lastcmd_lea_edi:=-1;
 lastcmd_add_edi:=Code.Size;
 lastcmd_add_edi_off:=si.Offset;
end
else
begin
 WriteData(#$8D#$BD+ConvertDWord(si.Offset+lastcmd_lea_edi_off)); // lea edi,[si.offset]
 lastcmd_lea_edi_off:=si.Offset+lastcmd_lea_edi_off;
end;
end
else
begin
 if si.Parent.BaseType=btclass then
 begin
  WriteData(#$89#$3F); // mov edi,[edi]
  if si.Offset<>0 then WriteData(#$81#$C7+ConvertDWord(si.Offset)); // add edi, si.value
  lastcmd_add_edi:=Code.Size;
  lastcmd_add_edi_off:=si.Offset;
 end
 else
 begin
  if lastcmd_add_edi=code.Size then
  begin
   Code.Seek(-6,soFromCurrent);
   lastcmd_add_edi_off:=lastcmd_add_edi_off+si.Offset;
   WriteData(#$81#$C7+ConvertDWord(lastcmd_add_edi_off)); // add edi, si.value
  end
  else
  begin
   if si.Offset<>0 then WriteData(#$81#$C7+ConvertDWord(si.Offset)); // add edi, si.value
   lastcmd_add_edi:=Code.Size;
   lastcmd_add_edi_off:=si.Offset;
  end;
 end;
end;
end;

procedure TCodeGenerator.LoadIntVar2Float(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$85+ConvertDWord(si.Offset)); // mov eax,[si.offset]
 WriteData(#$DB#$00); // fild [eax]
end
else
begin
 WriteData(#$DB#$85+ConvertDWord(si.Offset)); // fild [si.offset]
end;
IncFPUStack;
end;

procedure TCodeGenerator.LoadIntRefEDI;
begin
if (lastcmd_lea_edi=code.size) then
begin
code.Seek(-6,soFromCurrent);
WriteData(#$8B#$85+ConvertDWord(lastcmd_lea_edi_off)); // mov eax,[ebp+$FFFFF]
lastcmd_lea_edi:=-1;
end
else
 WriteData(#$8B#$07); // mov eax,[edi]
end;

procedure TCodeGenerator.LoadByteRefEDI;
begin
if (lastcmd_lea_edi=code.size) then
begin
code.Seek(-6,soFromCurrent);
WriteData(#$0F#$B6#$85+ConvertDWord(lastcmd_lea_edi_off)); // movzx eax,[ebp+$FFFFF]
lastcmd_lea_edi:=-1;
end
else
 WriteData(#$0F#$B6#$07); // mov eax,[edi]
end;

procedure TCodeGenerator.LoadFloatRefEDI;
begin
if (lastcmd_lea_edi=code.size) then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$D9#$85+ConvertDWord(lastcmd_lea_edi_off)); // fld [ebp+$FFFFF]
 lastcmd_lea_edi:=-1;
end
else
 WriteData(#$D9#$07); // fld [edi]
IncFPUStack;
end;

procedure TCodeGenerator.StoreIntRefEDI;
begin
if lastcmd_lea_edi=code.size then
begin
  code.Seek(-6,soFromCurrent);
 WriteData(#$89#$95+ConvertDWord(lastcmd_lea_edi_off)); // mov [ebp+$FFFF],edx
 lastcmd_lea_edi:=-1;
end
else
 WriteData(#$89#$17); // mov [edi],edx
FreeEDX;
end;

procedure TCodeGenerator.StoreIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$89#$01); // mov [ecx],eax
end
else
 WriteData(#$89#$85+ConvertDWord(si.Offset)); // mov [ebp+si.offset],eax
end;

procedure TCodeGenerator.StoreIntLocalConst(si:TSymbolInfo;c:integer);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$C1#$01+ConvertDWord(c)); // mov [ecx],c
end
else
 WriteData(#$C7#$85+ConvertDWord(si.Offset)+ConvertDWord(c)); // mov [ebp+si.offset],c
end;

procedure TCodeGenerator.StoreByteRefEDI;
begin
if lastcmd_lea_edi=code.size then
begin
  code.Seek(-6,soFromCurrent);
 WriteData(#$88#$95+ConvertDWord(lastcmd_lea_edi_off)); // mov [ebp+$FFFF],dl
 lastcmd_lea_edi:=-1;
end
else
 WriteData(#$88#$17); // mov [edi],dl
FreeEDX;
end;

procedure TCodeGenerator.StoreByteLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$88#$01); // mov [ecx],al
end
else
 WriteData(#$88#$85+ConvertDWord(si.Offset)); // mov [ebp+si.offset],al
end;

procedure TCodeGenerator.StoreByteLocalConst(si:TSymbolInfo;b:boolean);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$C6#$01+chr(Byte(b))); // mov [ecx],c
end
else
 WriteData(#$C6#$85+ConvertDWord(si.Offset)+chr(byte(b))); // mov [ebp+si.offset],c
end;


procedure TCodeGenerator.StoreFloatRefEDI;
begin
WriteData(#$D9#$1F); // fstp [edi]
DecFPUStack;
end;

procedure TCodeGenerator.StoreFloatLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$D9#$19); // fstp [ecx]
end
else
 WriteData(#$D9#$9D+ConvertDWord(si.Offset)); // fstp [ebp+si.offset]
end;

procedure TCodeGenerator.StoreFloatLocalConst(si:TSymbolInfo;c:single);
begin
 StoreIntLocalConst(si,PInteger(@c)^);
end;

procedure TCodeGenerator.StoreFloatRefIntEDI;
begin
WriteData(#$89#$54#$24#$FC); // mov [esp-4],eax
WriteData(#$D9#$44#$24#$FC); // fld [esp-4]
WriteData(#$D9#$1F); // fstp [edi]
DecFPUStack;
end;

procedure TCodeGenerator.MoveBlock(size:integer);
begin
if size mod 4=0 then
begin
 WriteData(#$B9+ConvertDWord(size div 4)); // mov ecx,size
 WriteData(#$F3#$A5); // rep movsd
end
else
begin
 WriteData(#$B9+ConvertDWord(size)); // mov ecx,size
 WriteData(#$F3#$A4); // rep movsb
end;
FreeESI;
end;

procedure TCodeGenerator.IntToFloat;
begin
WriteData(#$89#$44#$24#$FC); // mov [esp-4],eax
WriteData(#$DF#$44#$24#$FC); // fild [esp-4]
IncFPUStack;
end;

procedure TCodeGenerator.CallMethod(obj:boolean;si:TSymbolInfo);
begin
if (si.TypeInfo.MethodType in [mtconstructor,mtdestructor]) then
begin
 if obj then WriteData(#$B2#$00) else WriteData(#$B2#$01); // mov dl,0 ; mov dl,1
end;
if obj and (si.TypeInfo.MethodType in [mtclassfunction,mtclassprocedure]) then
begin
 WriteData(#$8B#$00); // mov eax,[eax]
 obj:=false;
end;

if obj then
begin
 if sfvirtual in si.Flags then
 begin
  WriteData(#$8B#$08); // mov ecx,[eax]
  WriteData(#$FF#$91+ConvertDWord(si.VMTIndex*4));// call [ecx+vmtindex*4]
 end
 else
   CallAddr(@si.MethodPtr);
end
else
begin
 if sfvirtual in si.Flags then
 begin
  WriteData(#$FF#$91+ConvertDWord(si.VMTIndex*4));// call [ecx+vmtindex*4]
 end
 else
   CallAddr(@si.MethodPtr);
end;
end;

procedure TCodeGenerator.AddIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
WriteData(#$03#$01); // add eax,[ecx]
end
else
WriteData(#$03#$85+ConvertDWord(si.Offset)); // add eax,[ebp+si.offset]
end;

procedure TCodeGenerator.AddIntConst(c:integer);
begin
WriteData(#$05+ConvertDWord(c)); // add eax,c
end;

procedure TCodeGenerator.AddIntValue;
begin
WriteData(#$01#$D0); // add eax,edx
FreeEDX;
end;

procedure TCodeGenerator.AddFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DA#$01); // fiadd [ecx]
  end
  else
   WriteData(#$DA#$85+ConvertDWord(si.Offset)); // fiadd dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$01); // fadd [ecx]
  end
  else
   WriteData(#$D8#$85+ConvertDWord(si.Offset)); // fadd dword [ebp+si.offset]
 end;
end;
end;

procedure TCodeGenerator.AddFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$44#$24#$FC); // fadd dword [esp-4]
end;

procedure TCodeGenerator.AddFloatValue;
begin
WriteData(#$DE#$C1); // faddp
DecFPUStack;
end;

procedure TCodeGenerator.AddStringValue(target:TSymbolInfo);
begin
WriteData(#$8D#$8D+ConvertDWord(target.Offset)); // lea ecx,[ebo+target.offset]
if edx_count>0 then PushReg(regedx);
WriteData(#$89#$F0); // mov eax,esi
WriteData(#$89#$FA); // mov edx,edi
//PushReg(regecx);
//PushReg(regedi);
//PushReg(regesi);
CallAddr(@System_StringConcat);
if edx_count>0 then PopReg(regedx);
FreeESI;
LoadVar(target);
end;

procedure TCodeGenerator.IncVar(si:TSymbolInfo);
begin
if sfreference in si.flags then
begin
 LoadVarAddress(si);
 WriteData(#$FF#$07); // dec [edi]
end
else
 WriteData(#$FF#$85+ConvertDWord(si.Offset)); // dec [ebp+si.offset]
end;


procedure TCodeGenerator.SubIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$2B#$01); // sub eax,[ecx]
end
else
 WriteData(#$2B#$85+ConvertDWord(si.Offset)); // sub eax,[ebp+si.offset]
end;

procedure TCodeGenerator.SubIntConst(c:integer);
begin
WriteData(#$2D+ConvertDWord(c)); // sub eax,c
end;

procedure TCodeGenerator.SubIntValue;
begin
WriteData(#$29#$D0); // sub eax,edx
FreeEDX;
end;

procedure TCodeGenerator.SubFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DA#$21); // fisub [ecx]
  end
  else
   WriteData(#$DA#$A5+ConvertDWord(si.Offset)); // fisub dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$21); // fsub [edx]
  end
  else
   WriteData(#$D8#$A5+ConvertDWord(si.Offset)); // fsub dword [ebp+si.offset]
 end;
end;
end;

procedure TCodeGenerator.SubFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$64#$24#$FC); // fsubr dword [esp-4]
end;

procedure TCodeGenerator.SubFloatValue;
begin
WriteData(#$DE#$E9); // fsubrp
DecFPUStack;
end;

procedure TCodeGenerator.DecEBX;
begin
 WriteData(#$4B); // dec ebx
end;

procedure TCodeGenerator.DecVar(si:TSymbolInfo);
begin
if sfreference in si.flags then
begin
 LoadVarAddress(si);
 WriteData(#$FF#$0F); // dec [edi]
end
else
 WriteData(#$FF#$8D+ConvertDWord(si.Offset)); // dec [ebp+si.offset]
end;

procedure TCodeGenerator.MultIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
WriteData(#$F7#$29); // imul [ecx]
end
else
WriteData(#$F7#$AD+ConvertDWord(si.Offset)); // imul eax,[ebp+si.offset]
end;

procedure TCodeGenerator.MultIntConst(c:Integer);
begin
WriteData(#$69#$C0+ConvertDWord(c)); // imul eax,c
end;

procedure TCodeGenerator.MultIntValue;
begin
WriteData(#$F7#$EA); // imul edx
FreeEDX;
end;

procedure TCodeGenerator.MultFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DA#$09); // fimul [ecx]
  end
  else
   WriteData(#$DA#$8D+ConvertDWord(si.Offset)); // fimul dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$09); // fmul [edx]
  end
  else
   WriteData(#$D8#$8D+ConvertDWord(si.Offset)); // fmul dword [ebp+si.offset]
 end;
end;
end;

procedure TCodeGenerator.MultFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$4C#$24#$FC); // fmul dword [esp-4]
end;

procedure TCodeGenerator.MultFloatValue;
begin
WriteData(#$DE#$C9); // fmulp
DecFPUStack;
end;

procedure TCodeGenerator.DivIntLocal(si:TSymbolInfo);
begin
if edx_count>0 then PushReg(regedx);
WriteData(#$31#$D2); // xor edx,edx 
if sfreference in si.Flags then
begin
WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
WriteData(#$F7#$39); // idiv [ecx]
end
else
WriteData(#$F7#$BD+ConvertDWord(si.Offset)); // idiv eax,[ebp+si.offset]
if edx_count>0 then PopReg(regedx);
end;

procedure TCodeGenerator.DivIntConst(c:Integer);
begin
if edx_count>0 then PushReg(regedx);
WriteData(#$31#$D2); // xor edx,edx
 WriteData(#$B9+ConvertDWord(c)); // mov ecx,c
WriteData(#$F7#$F9); // idiv ecx
if edx_count>0 then PopReg(regedx);
end;

procedure TCodeGenerator.DivIntValue;
begin
WriteData(#$89#$CA); // mov ecx,edx
WriteData(#$31#$D2); // xor edx,edx
WriteData(#$F7#$F9); // idiv ecx
FreeEDX;
end;

procedure TCodeGenerator.DivFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DA#$31); // fidivr [ecx]
  end
  else
   WriteData(#$DA#$B5+ConvertDWord(si.Offset)); // fidivr dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$31); // fdivr [edx]
  end
  else
   WriteData(#$D8#$B5+ConvertDWord(si.Offset)); // fdivr dword [ebp+si.offset]
 end;
end;
end;

procedure TCodeGenerator.DivFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$74#$24#$FC); // fdivr dword [esp-4]
end;

procedure TCodeGenerator.DivFloatValue;
begin
WriteData(#$DE#$F9); // fdivrp
DecFPUStack;
end;

procedure TCodeGenerator.ModIntLocal(si:TSymbolInfo);
begin
if edx_count>0 then PushReg(regedx);
WriteData(#$31#$D2); // xor edx,edx
if sfreference in si.Flags then
begin
WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
WriteData(#$F7#$39); // idiv [ecx]
end
else
WriteData(#$F7#$BD+ConvertDWord(si.Offset)); // idiv eax,[ebp+si.offset]
WriteData(#$89#$D0); // mov eax,edx
if edx_count>0 then PopReg(regedx);
end;

procedure TCodeGenerator.ModIntConst(c:Integer);
begin
if edx_count>0 then PushReg(regedx);
WriteData(#$31#$D2); // xor edx,edx
 WriteData(#$B9+ConvertDWord(c)); // mov ecx,c
WriteData(#$F7#$F9); // idiv ecx
WriteData(#$89#$D0); // mov eax,edx
if edx_count>0 then PopReg(regedx);
end;

procedure TCodeGenerator.ModIntValue;
begin
WriteData(#$89#$CA); // mov ecx,edx
WriteData(#$31#$D2); // xor edx,edx
WriteData(#$F7#$F9); // idiv ecx
WriteData(#$89#$D0); // mov eax,edx
FreeEDX;
end;

procedure TCodeGenerator.EqualIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$94#$C0); // setz al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_z:=code.size;
end;

procedure TCodeGenerator.EqualStringValue;
begin
//WriteData(#$8D#$8D+ConvertDWord(target.Offset)); // lea ecx,[ebo+target.offset]
if edx_count>0 then PushReg(regedx);
WriteData(#$89#$F0); // mov eax,esi
WriteData(#$89#$FA); // mov edx,edi
CallAddr(@System_StringEqual);
if edx_count>0 then PopReg(regedx);
FreeESI;
end;

procedure TCodeGenerator.NotEqualIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$95#$C0); // setnz al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_nz:=code.size;
end;

procedure TCodeGenerator.NotEqualStringValue;
begin
//WriteData(#$8D#$8D+ConvertDWord(target.Offset)); // lea ecx,[ebo+target.offset]
if edx_count>0 then PushReg(regedx);
WriteData(#$89#$F0); // mov eax,esi
WriteData(#$89#$FA); // mov edx,edi
CallAddr(@System_StringEqual);
if edx_count>0 then PopReg(regedx);
FreeESI;
WriteData(#$34#$01); // xor al,01
end;

procedure TCodeGenerator.NotBoolean;
begin
WriteData(#$34#$01); // xor al,01
end;

procedure TCodeGenerator.NegInteger;
begin
WriteData(#$F7#$D8); // fchs
end;

procedure TCodeGenerator.NegFloat;
begin
WriteData(#$D9#$E0); // fchs
end;

procedure TCodeGenerator.GreaterIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9F#$C0); // setg al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_g:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.GreaterEqualFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9D#$C0); // setge al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_ge:=code.Size;
end;

procedure TCodeGenerator.LessIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9C#$C0); // setl al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_l:=code.Size;
end;

procedure TCodeGenerator.LessEqualIntLocal(si:TSymbolInfo);
begin
if si.TypeInfo.Size=4 then
 WriteData(#$3B#$85+ConvertDWord(si.Offset)) // cmp eax,[ebp+si.offset]
else
 WriteData(#$3A#$85+ConvertDWord(si.Offset)); // cmp al,[ebp+si.offset];
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.LessEqualIntConst(c:Integer);
begin
WriteData(#$3D+ConvertDWord(c)); // cmp eax,edx
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.LessEqualIntValue;
begin
WriteData(#$39#$D0); // cmp eax,edx
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.LessEqualFloatLocal(si:TSymbolInfo);
begin
case si.typeinfo.basetype of
 btinteger:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$DE#$19); // ficomp [ecx]
  end
  else
   WriteData(#$DE#$9D+ConvertDWord(si.Offset)); // ficomp dword [ebp+si.offset]
 end;
 btsingle:
 begin
  if sfreference in si.Flags then
  begin
   WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
   WriteData(#$D8#$19); // fcomp [ecx]
  end
  else
   WriteData(#$D8#$9D+ConvertDWord(si.Offset)); // fcomp dword [ebp+si.offset]
 end;
end;


WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.LessEqualFloatConst(s:single);
begin
WriteData(#$C7#$44#$24#$FC+ConvertDWord(PInteger(@s)^)); // mov [esp-4],f
WriteData(#$D8#$5C#$24#$FC); // fcomp dword [esp-4]
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
DecFPUStack;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.LessEqualFloatValue;
begin
WriteData(#$DE#$D9); // fcompp
WriteData(#$DF#$E0); // fstsw ax
WriteData(#$9E); // sahf
WriteData(#$0F#$9E#$C0); // setle al
Movzx_EAX_AL;
DecFPUStack;
DecFPUStack;
lastcmd_setal_le:=code.Size;
end;

procedure TCodeGenerator.AndIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$23#$01); // and eax,[ecx]
end
else
 WriteData(#$23#$85+ConvertDWord(si.Offset)); // and eax,[ebp+si.offset]
end;

procedure TCodeGenerator.AndBooleanLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$22#$01); // and eax,[ecx]
end
else
 WriteData(#$22#$85+ConvertDWord(si.Offset)); // and eax,[ebp+si.offset]
end;

procedure TCodeGenerator.AndConst(c:integer);
begin
 WriteData(#$25+ConvertDWord(c)); // and eax,c
end;

procedure TCodeGenerator.AndValue;
begin
WriteData(#$21#$D0); // and eax,edx
Movzx_EAX_AL;
end;

procedure TCodeGenerator.OrIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$0B#$01); // or eax,[ecx]
end
else
 WriteData(#$0B#$85+ConvertDWord(si.Offset)); // or eax,[ebp+si.offset]
end;

procedure TCodeGenerator.OrBooleanLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$0A#$01); // or al,[ecx]
end
else
 WriteData(#$0A#$85+ConvertDWord(si.Offset)); // or al,[ebp+si.offset]
end;

procedure TCodeGenerator.OrConst(c:integer);
begin
 WriteData(#$0D+ConvertDWord(c)); // or eax,c
end;

procedure TCodeGenerator.OrValue;
begin
WriteData(#$09#$D0); // or eax,edx
Movzx_EAX_AL;
end;

procedure TCodeGenerator.XorIntLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$33#$01); // xor eax,[ecx]
end
else
 WriteData(#$33#$85+ConvertDWord(si.Offset)); // xor eax,[ebp+si.offset]
end;

procedure TCodeGenerator.XorBooleanLocal(si:TSymbolInfo);
begin
if sfreference in si.Flags then
begin
 WriteData(#$8B#$8D+ConvertDWord(si.Offset)); // mov ecx,[ebp+si.offset]
 WriteData(#$32#$01); // xor al,[ecx]
end
else
 WriteData(#$32#$85+ConvertDWord(si.Offset)); // xor al,[ebp+si.offset]
end;

procedure TCodeGenerator.XorConst(c:Integer);
begin
 WriteData(#$35+ConvertDWord(c)); // or eax,c
end;

procedure TCodeGenerator.XorValue;
begin
WriteData(#$31#$D0); // or eax,edx
Movzx_EAX_AL;
end;


function TCodeGenerator.JumpTrue:integer;
begin
WriteData(#$84#$C0); // test al,al
WriteData(#$74#$05); // jz +5
result:=code.Size;
WriteData(#$E9+ConvertDWord(0)); // jmp ...
end;

function TCodeGenerator.JumpFalse:integer;
begin
if lastcmd_setal_z=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$74#$05); // jz +5
 lastcmd_setal_z:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
if lastcmd_setal_nz=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$75#$05); // jnz +5
 lastcmd_setal_nz:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
if lastcmd_setal_g=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$7F#$05); // jg +5
 lastcmd_setal_g:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
if lastcmd_setal_ge=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$7D#$05); // jge +5
 lastcmd_setal_ge:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
if lastcmd_setal_l=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$7C#$05); // jl +5
 lastcmd_setal_l:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
if lastcmd_setal_g=code.size then
begin
 code.Seek(-6,soFromCurrent);
 WriteData(#$7E#$05); // jle +5
 lastcmd_setal_le:=-1;
 result:=code.Size-4;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end
else
begin
 WriteData(#$84#$C0); // test al,al
 WriteData(#$75#$05); // jnz +5
 result:=code.Size;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end;

end;

function TCodeGenerator.Jump_Zero:Integer;
begin
 WriteData(#$75#$05); // jnz +5
 result:=code.Size;
 WriteData(#$E9+ConvertDWord(0)); // jmp ...
end;

function TCodeGenerator.Jump:integer;
begin
result:=code.Size;
WriteData(#$E9+ConvertDWord(0)); // jmp ...
end;

procedure TCodeGenerator.SetJumpTarget(jumpop,target:Integer);
var
 p:integer;
begin
p:=code.Position;
code.Seek(jumpop+1,soFromBeginning);
WriteData(ConvertDWord(target-jumpop-5));
code.Seek(p,soFromBeginning);
end;


constructor TCodeGenerator.create;
begin
 inherited;
 flocalvars:=THashedStringList.Create;
 flocalvars.CaseSensitive:=false;
end;

destructor TCodeGenerator.destroy;
begin
 code.Free;
 flocalvars.free;
 inherited;
end;

// TUnitInfo

function TUnitInfo.AddClass(const cname:string;baseclass:TTypeInfo):TTypeInfo;
begin
if IsDeclared(cname) then
 raise ESyntaxError.Create('Identifier redeclared: "'+cname+'"');
result:=TTypeInfo.create(btclass,baseclass);
result.funitinfo:=self;
result.fname:=cname;
result.fbasetype:=btclass;
result.InterfaceDecl:=InterfaceSec;
ftypelist.AddObject(cname,result);
result.fclasstypeinfo:=result.CreateClassRefType;
end;

function TUnitInfo.AddType(const tname:string;basetype:TBaseType):TTypeInfo;
begin
if IsDeclared(tname) then
 raise ESyntaxError.Create('Identifier redeclared: "'+tname+'"');
result:=TTypeInfo.create(btclass,nil);
result.fname:=tname;
result.funitinfo:=self;
result.fbasetype:=basetype;
result.InterfaceDecl:=InterfaceSec;
ftypelist.AddObject(tname,result);
end;

function TUnitInfo.AddRecord(const cname:string;basetype:TTypeInfo):TTypeInfo;
begin
if IsDeclared(cname) then
 raise ESyntaxError.Create('Identifier redeclared: "'+cname+'"');
result:=TTypeInfo.create(btrecord,basetype);
result.funitinfo:=self;
result.fname:=cname;
result.InterfaceDecl:=InterfaceSec;
ftypelist.AddObject(cname,result);
end;

function TUnitInfo.AddEnum(const tname:string;const enumlist:array of string):TTypeInfo;
var
 i:integer;
begin
result:=AddType(tname,btenum);
for i:=low(enumlist) to high(enumlist) do
begin
 AddConst(enumlist[i],inttostr(i),result);
 result.AddField(enumlist[i],System_Integer,fspublic);
end;

end;

procedure TUnitInfo.AddImport(unitinfo:TUnitInfo);
begin
interfaceimplist.Add(unitinfo);
end;

function TUnitInfo.AddSymbol(const name:string;t:TTypeInfo):TSymbolInfo;
begin
if IsDeclared(name) then Redeclared(name);
result:=TSymbolInfo.create(name,t);
result.fUnitInfo:=self;
result.finterfacedecl:=InterfaceSec;
fsymbollist.AddObject(name,result);
end;

function TUnitInfo.AddFunction(const name:string;mt:TMethodType):TSymbolInfo;
var
 t:TTypeInfo;
begin
t:=AddType(self.Name+'.'+name,btmethod);
t.MethodType:=mt;
result:=AddSymbol(name,t);
end;

function TUnitInfo.AddConst(const name,val:string;t:TTypeInfo):TSymbolInfo;
begin
result:=AddSymbol(name,t);
include(result.fflags,sftrueconst);
result.fconstdata:=val;
end;

function TUnitInfo.AddStringConst(const name:string):Integer;
var
 p:pointer;
begin
getmem(p,length(name)+1);
PShortString(p)^:=name;
stringconstlist.add(p);
result:=integer(p);
end;

function TUnitInfo.FindType(const name:string;noimp:boolean=false):TTypeInfo;
var
 i:integer;
 unitinfo:TUnitInfo;
 s,s2:string;
 found:boolean;
begin
result:=nil;
i:=pos('.',name);
if i=0 then
begin
i:=ftypelist.IndexOf(name);
if i=-1 then result:=nil else result:=TTypeInfo(ftypelist.objects[i]);
if assigned(result) and not result.InterfaceDecl and noimp then result:=nil;
if not noimp and (result=nil) then
begin
 for i:=interfaceimplist.count-1 downto 0 do
 begin
  result:=TUnitInfo(interfaceimplist[i]).FindType(name,true);
  if assigned(result) then break;
 end;
end;
end
else
begin
s:=ExtractFileExt(name);
s:=copy(s,2,length(s)-1);
s2:=copy(name,1,length(name)-length(s)-1);
found:=SameText(self.Name,s2);
if not found and noimp then exit;
for i:=0 to interfaceimplist.count-1 do
 if sametext(s2,TUnitInfo(interfaceimplist[i]).Name) then
 begin
  found:=true;
  break;
 end;
if not found then exit;
unitinfo:=FindUnit(s2);
if unitinfo=nil then exit;
result:=unitinfo.FindType(s,true);
if assigned(result) and not result.InterfaceDecl and noimp then result:=nil;
end;
end;

function TUnitInfo.FindSymbol(const name:string;noimp:boolean=false):TSymbolInfo;
var
 i:integer;
 unitinfo:TUnitInfo;
 s,s2:string;
 found:boolean;
begin
result:=nil;
i:=pos('.',name);
if i=0 then
begin
i:=fsymbollist.IndexOf(name);
if i=-1 then result:=nil else result:=TSymbolInfo(fsymbollist.objects[i]);
if assigned(result) and not result.InterfaceDecl and noimp then result:=nil;
if not noimp and (result=nil) then
begin
 for i:=interfaceimplist.count-1 downto 0 do
 begin
  result:=TUnitInfo(interfaceimplist[i]).FindSymbol(name,true);
  if assigned(result) then break;
 end;
end;
end
else
begin
s:=ExtractFileExt(name);
s:=copy(s,2,length(s)-1);
s2:=copy(name,1,length(name)-length(s)-1);
found:=SameText(self.Name,s2);
if not found and noimp then exit;
for i:=0 to interfaceimplist.count-1 do
 if sametext(s2,TUnitInfo(interfaceimplist[i]).Name) then
 begin
  found:=true;
  break;
 end;
if not found then exit;
unitinfo:=FindUnit(s2);
if unitinfo=nil then exit;
result:=unitinfo.FindSymbol(s,true);
if assigned(result) and not result.InterfaceDecl and noimp then result:=nil;
end;
end;

function TUnitInfo.IsDeclared(const name:string):boolean;
begin
result:=(ftypelist.IndexOf(name)<>-1) or (fsymbollist.IndexOf(name)<>-1);
end;

class function TUnitInfo.FindUnit(const name:string):TUnitInfo;
var
 i:integer;
begin
result:=nil;
for i:=0 to unitlist.count-1 do
 if SameText(TUnitInfo(unitlist[i]).Name,name) then
 begin
  result:=TUnitInfo(unitlist[i]);
  break;
 end;
end;

function TUnitInfo.CheckImplemention:TSymbolInfo;
var
 i,j:integer;
begin
result:=nil;
for i:=0 to ftypelist.count-1 do
 if TTypeInfo(ftypelist.objects[i]).BaseType=btclass then
  for j:=0 to TTypeInfo(ftypelist.objects[i]).FieldCount-1 do
   if TTypeInfo(ftypelist.objects[i]).Fields[j].TypeInfo.BaseType=btmethod then
    if not TTypeInfo(ftypelist.objects[i]).Fields[j].Implemented then
    begin
     result:=TTypeInfo(ftypelist.objects[i]).Fields[j];
     exit;
    end;
for i:=0 to fsymbollist.count-1 do
 with TSymbolInfo(fsymbollist.objects[i]) do
  if (TypeInfo.BaseType=btmethod) and not Implemented then
  begin
   result:=TSymbolInfo(fsymbollist.objects[i]);
   exit;
  end;
end;

procedure TUnitInfo.RegisterClasses;
var
 i:integer;
 c:TClass;
begin
for i:=0 to ftypelist.count-1 do
 begin
  if TTypeInfo(ftypelist.objects[i]).BaseType=btclass then
  begin
   TTypeInfo(ftypelist.objects[i]).CreateClass;
   TTypeInfo(ftypelist.objects[i]).CreateRTTInfo;
   c:=TTypeInfo(ftypelist.objects[i]).ClassPtr;
   RegisterClass(TPersistentClass(c));
  end
  else
   TTypeInfo(ftypelist.objects[i]).CreateRTTInfo;
 end;

end;

procedure TUnitInfo.UnRegisterClasses;
var
 i:integer;
begin
for i:=0 to ftypelist.count-1 do
 with TTypeInfo(ftypelist.objects[i]) do
  if unregister then
   UnRegisterClass(TPersistentClass(ClassPtr));
end;

constructor TUnitInfo.create(const name:string);
begin
 inherited create;
 fname:=name;
 ftypelist:=THashedStringList.Create;
 fsymbollist:=THashedStringList.Create;
 stringconstlist:=TList.create;
 interfaceimplist:=TList.create;
 unitlist.Add(self);
 if not SameText(name,'System') then AddImport(FindUnit('System'));
end;

destructor TUnitInfo.destroy;
var
 i:integer;
begin
 UnRegisterClasses;
 for i:=0 to stringconstlist.count-1 do
  FreeMem(stringconstlist[i]);
 stringconstlist.Free; 
 for i:=0 to fsymbollist.Count-1 do
  fsymbollist.Objects[i].Free;
 fsymbollist.Free;
 for i:=0 to ftypelist.Count-1 do
  ftypelist.Objects[i].Free;
 ftypelist.Free;

 if assigned(unitlist) then unitlist.Extract(self);
 interfaceimplist.Free;
 inherited;
end;


// esi: source edi:destination
procedure StringCopy;
asm
 movzx ecx,[esi];
 add ecx,1;
 rep movsb;
 ret;
end;


function StringConcat(const a,b:shortstring):ShortString;
begin
result:=a+b;
end;

function StringEqual(const a,b:shortstring):boolean;
begin
result:=a=b;
end;

procedure InitSystemUnit;
var
 cls:TTypeInfo;
begin
with TUnitInfo.create('System') do
begin
 InterfaceSec:=true;
 System_Integer:=AddType('Integer',btinteger);
 System_Integer.fsize:=4;
 System_Integer.frttinfo:=GetPropInfo(TTypeInfoClass,'IntProp').PropType;
 System_Untyped:=AddType(' Untyped',btinteger);
 System_Untyped.fsize:=4;
 System_Single:=AddType('Single',btsingle);
 System_Single.fsize:=4;
 System_Single.frttinfo:=GetPropInfo(TTypeInfoClass,'SingleProp').PropType;
 System_Boolean:=AddType('Boolean',btboolean);
 System_Boolean.fSize:=1;
 System_Boolean.frttinfo:=GetPropInfo(TTypeInfoClass,'BooleanProp').PropType;
 AddConst('false','0',System_Boolean);
 AddConst('true','1',System_Boolean);
 System_Nil:=AddType('System.Nil',btnil);
 System_Nil.fsize:=4;
 System_String:=AddType('String',btstring);
 System_String.fsize:=256;
 System_String.frttinfo:=GetPropInfo(TTypeInfoClass,'StrProp').PropType;
 cls:=AddClass('TObject',nil);
 cls.fclassptr:=TPersistent;
 cls.AddMethod('Create',mtconstructor,fspublic,@TObject.Create);
 cls.AddMethod('Free',mtprocedure,fspublic,@TObject.Free);
 cls.fsize:=4;
 cls.finstsize:=4;
 cls.vmtsize:=3;
 with cls.AddMethod('Destroy',mtdestructor,fspublic,@TObject.Destroy) do
 begin
  Flags:=Flags+[sfvirtual];
  fvmtindex:=-1;
 end;
 with cls.AddMethod('Dispatch',mtprocedure,fspublic,@TObject.Dispatch) do
 begin
  Typeinfo.AddParam('Message',System_Untyped,ptvar);
  Flags:=Flags+[sfvirtual];
  fvmtindex:=-5;
 end;
 System_Procedure:=AddType('procedure',btprocedure);
end;
System_StringCopy:=@StringCopy;
System_StringConcat:=@StringConcat;
System_StringEqual:=@StringEqual;
end;

procedure InitFuncPtr;
var
 p:pointer;
begin
p:=pointer(integer(@TObject.Create)+8);
_classcreate:=Pointer(PInteger(p)^+Integer(p)+4);
p:=pointer(integer(@TObject.Create)+17);
_afterconstruction:=Pointer(PInteger(p)^+Integer(p)+4);
p:=pointer(integer(@TObject.Destroy)+1);
_beforedestruction:=Pointer(PInteger(p)^+Integer(p)+4);
p:=pointer(integer(@TObject.Destroy)+10);
_classdestroy:=Pointer(PInteger(p)^+Integer(p)+4);
//asm
// call _classdestroy;
//end;
end;

initialization
begin
 InitFuncPtr;
 unitlist:=TObjectList.Create;
 InitSystemUnit;
end;

finalization
begin
 System_Integer.UnitInfo.free;
 unitlist.OwnsObjects:=false;
 FreeAndNil(unitlist);
end;

end.
