unit dp_scanner;

// Dynamic Pascal, Copyright Lars Middendorf

interface

uses classes,sysutils;

type


TTokenType=(tkident,tkinteger,tkfloat,tkstring,tkeof,tkerror,
            tkprogram,tkunit,tkpackage,tkinterface,tkimplementation,tkuses,tkrequires,tkcontains,tkprocedure,tkfunction,tkconstructor,tkdestructor,tkforward,tkexternal,tkexports,tkbegin,tkend,tkif,tkthen,tkelse,tkwhile,tkfor,tkto,tkdo,tkrepeat,tkuntil,tkvar,tkconst,tktype,tkrecord,tkclass,tkarray,tkof,tkidiv,tkmod,tkshr,tkshl,tkand,tkor,tkxor,tknot,tkprivate,tkprotected,tkpublic,tkpublished,tkvirtual,tkoverride,tkmessage,tknil,tkbreak,tkexit,tkcontinue,
            tkassignment,tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkaddress,tkderef,tksemicolon,tkcomma,tkopenbracket,tkclosebracket,tkpoint,tkdblpoint,tkarrayopen,tkarrayclose,tkrange,
            tktypecast,tkfunccall,tklast);
const

tokentext:array[tkident..tklast] of string =('','','','','','',
            'program','unit','package','interface','implementation','uses','requires','contains','procedure','function','constructor','destructor','forward','external','exports','begin','end','if','then','else','while','for','to','do','repeat','until','var','const','type','record','class','array','of','div','mod','shr','shl','and','or','xor','not','private','protected','public','published','virtual','override','message','nil','break','exit','continue',
            ':=','+','-','*','/','=','<=','<','>=','>','<>','@','^',';',',','(',')','.',':','[',']','..',
            '','','');

binaryop=[tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkpoint,tkidiv,tkmod,tkshr,tkshl,tkand,tkor,tkxor];
unaryop=[tkadd,tksub,tknot,tkaddress];
postop=[tkderef];
expressionstart=[tkident,tkinteger,tkfloat,tkstring]+unaryop;

type
TTokenSet=set of TTokenType;

ESyntaxError=class(Exception);

TScanner=class
private
 current_char,next_char:char;
 stream:TStream;
 function ReadChar:char;
 procedure ReadText;
 procedure FindTokenStart;
 procedure ReadSymbol;
 procedure ReadNum;
 function ReadString:string;
public
 SourceLine:integer;
 SourcePos:integer;
 TokenString:string;
 TokenType:TTokenType;
 function LineStr:String;
 function ReadToken:string;
 constructor create(st:TStream);
 destructor destroy;override;
end;

procedure SyntaxError(const msg:string);
procedure Undeclared(const msg:string);
procedure Redeclared(const msg:string);

implementation

var
 currentsc:TScanner;

procedure SyntaxError(const msg:string);
begin
 raise ESyntaxError.Create(currentsc.LineStr+' :'+msg);
end;

procedure Undeclared(const msg:string);
begin
SyntaxError('Undeclared identifier: '''+msg+'''');
end;

procedure Redeclared(const msg:string);
begin
SyntaxError('Identifier redeclared: '''+msg+'''');
end;


function TScanner.ReadChar:char;
begin
current_char:=next_char;
if stream.Position>=stream.Size then
 next_char:=#0
else
begin
 stream.Read(next_char,1);
 inc(SourcePos);
end;
if current_char=#10 then inc(sourceline);
result:=current_char;
end;

procedure TScanner.FindTokenStart;
begin
while current_char in [' ',#13,#10] do ReadChar;
if current_char='{' then
begin
 while not (ReadChar in ['}',#0]) do;
end;
end;

procedure TScanner.ReadText;
var
 t:TTokenType;
begin
TokenType:=tkident;
TokenString:=current_char;
while ReadChar in ['A'..'Z','a'..'z','_','0'..'9'] do TokenString:=TokenString+current_char;
for t:=tkident to tklast do
 if sametext(Tokentext[t],TokenString) then
 begin
  TokenType:=t;
  break;
 end;
end;

procedure TScanner.ReadSymbol;
begin
TokenString:=current_char;
TokenType:=tkerror;
case current_char of
 '+':TokenType:=tkadd;
 '-':TokenType:=tksub;
 '*':TokenType:=tkmult;
 '/':
 if next_char='/' then
 begin
  ReadChar;
  TokenString:='//';
  end
 else
  TokenType:=tkdiv;
 ';':TokenType:=tksemicolon;
 ',':TokenType:=tkcomma;
 '(':TokenType:=tkopenbracket;
 ')':TokenType:=tkclosebracket;
 '.':
  begin
   if next_char='.' then
   begin
    TokenType:=tkrange;
    ReadChar;
    TokenString:='..';
   end
   else
   begin
    TokenType:=tkpoint;
   end; 
  end;
 '=':TokenType:=tkequal;
 '@':TokenType:=tkaddress;
 '^':TokenType:=tkderef;
 '[':TokenType:=tkarrayopen;
 ']':TokenType:=tkarrayclose;
 '<':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='<=';
    TokenType:=tklequal;
   end
   else
   if next_char='>' then
   begin
    ReadChar;
    TokenString:='<>';
    TokenType:=tknotequal;
   end
   else
   begin
    TokenType:=tkless;
   end;
  end;
 '>':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='>=';
    TokenType:=tkgequal;
   end
   else
   begin
    TokenType:=tkgreater;
   end;
  end;
 ':':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:=':=';
    TokenType:=tkassignment;
   end
   else
   begin
    TokenType:=tkdblpoint;
   end;
  end;
end;
ReadChar;
end;

procedure TScanner.ReadNum;
begin
Tokenstring:=current_char;
TokenType:=tkinteger;
while (ReadChar in ['0'..'9','.']) do
begin
 if (current_char='.') and (next_char='.') then break;
 if (current_char='.') then
 begin
  if TokenType=tkinteger then tokentype:=tkfloat else tokentype:=tkerror;
 end;
 Tokenstring:=Tokenstring+current_char;
end;

end;

function TScanner.ReadString:string;
begin
while not (ReadChar in ['''',#13,#10,#0]) do TokenString:=TokenString+current_char;
if current_char='''' then TokenType:=tkstring else TokenType:=tkerror;
ReadChar;
end;

function TScanner.LineStr:String;
begin
result:='Line '+inttostr(SourceLine);
end;

function TScanner.ReadToken:string;
begin
currentsc:=self;
FindTokenStart;
TokenString:='';
TokenType:=tkerror;
case current_char of
 'A'..'Z','a'..'z','_':ReadText;
 '0'..'9':ReadNum;
 '''':ReadString;
 #0:
 begin
  result:=#0;
  TokenType:=tkeof;
 end
else
 ReadSymbol;
if TokenString='//' then
begin
while not (ReadChar in [#13,#10,#0]) do;
result:=ReadToken;
end;
end;
result:=TokenString;
end;


constructor TScanner.create(st:TStream);
begin
inherited create;
SourceLine:=1;
SourcePos:=2;
stream:=st;
stream.ReadBuffer(current_char,1);
stream.ReadBuffer(next_char,1);
ReadToken;
end;

destructor TScanner.destroy;
begin
inherited;
end;

begin
DecimalSeparator:='.';
end.
